use eframe::egui::{
    Align2, Color32, Context, Pos2, Rect, Rounding, Shape, Stroke, Ui, Vec2, Window,
};
use eframe::emath;
use eframe::epaint::RectShape;

#[derive(Clone)]
pub struct Grid {
    pub name: &'static str,
    pub offset_x: f32,
    pub offset_y: f32,
    pub size_cells: usize,
    pub size_pixels: f32,
    pub stroke: Color32,
    pub stroke_width: f32,
    pub background: Color32,
    fill: Vec<Color32>,
}

impl Default for Grid {
    fn default() -> Self {
        Self {
            name: "grid",
            offset_x: 0.0,
            offset_y: 0.0,
            size_cells: 16,
            size_pixels: 400.0,
            stroke: Color32::GRAY,
            stroke_width: 1.0,
            background: Color32::from_rgb(230, 230, 230),
            fill: Vec::new(),
        }
    }
}

impl Grid {
    fn index(&self, x: usize, y: usize) -> usize {
        assert!(x < self.size_cells);
        assert!(y < self.size_cells);
        x + y * self.size_cells
    }

    pub fn fill(&self, x: usize, y: usize) -> Color32 {
        self.fill
            .get(self.index(x, y))
            .cloned()
            .unwrap_or(Color32::TRANSPARENT)
    }

    pub fn set_fill(&mut self, x: usize, y: usize, fill: Color32) {
        let idx = self.index(x, y);
        loop {
            match self.fill.get_mut(idx) {
                Some(f) => {
                    *f = fill;
                    return;
                }
                None => self.fill.push(Color32::TRANSPARENT),
            }
        }
    }

    pub fn clear_fill(&mut self) {
        self.fill.clear();
    }

    pub fn show(&self, ctx: &Context) {
        Window::new(self.name)
            .title_bar(false)
            //.frame(Frame::none())
            .anchor(Align2::CENTER_CENTER, Vec2::ZERO)
            .fixed_size(Vec2::splat(self.size_pixels))
            .show(ctx, |ui| {
                let (_id, rect) = ui.allocate_space(Vec2::splat(self.size_pixels));
                let to_screen = emath::RectTransform::from_to(
                    Rect::from_x_y_ranges(0.0..=1.0, 0.0..=1.0),
                    rect,
                );

                if self.background.a() > 0 {
                    ui.painter().add(Shape::Rect(RectShape::filled(
                        to_screen.transform_rect(Rect::from_x_y_ranges(0f32..=1f32, 0f32..=1f32)),
                        Rounding::none(),
                        self.background,
                    )));
                }

                let cell_width = 1.0 / self.size_cells as f32;
                for y in 0..self.size_cells {
                    let y_coord = y as f32 / self.size_cells as f32;
                    for x in 0..self.size_cells {
                        let idx = self.index(x, y);
                        if let Some(fill) = self.fill.get(idx) {
                            if fill.a() > 0 {
                                let x_coord = x as f32 / self.size_cells as f32;
                                ui.painter().add(Shape::Rect(RectShape::filled(
                                    to_screen.transform_rect(Rect::from_x_y_ranges(
                                        x_coord..=x_coord + cell_width,
                                        y_coord..=y_coord + cell_width,
                                    )),
                                    Rounding::none(),
                                    *fill,
                                )));
                            }
                        }
                    }
                }

                if self.stroke.a() > 0 && self.stroke_width > 0.0 {
                    for c in 0..=self.size_cells {
                        let coord = c as f32 / self.size_cells as f32;
                        // Horizontal.
                        ui.painter().add(Shape::LineSegment {
                            points: [
                                to_screen * Pos2::new(0.0, coord),
                                to_screen * Pos2::new(1.0, coord),
                            ],
                            stroke: Stroke::new(self.stroke_width, self.stroke),
                        });
                        // Vertical.
                        ui.painter().add(Shape::LineSegment {
                            points: [
                                to_screen * Pos2::new(coord, 0.0),
                                to_screen * Pos2::new(coord, 1.0),
                            ],
                            stroke: Stroke::new(self.stroke_width, self.stroke),
                        });
                    }
                }
            });
    }
}
